"""
Main command group for Keops' CLI.

Subcommands developed as a part of the Keops package have their own
modules under ``keops`` (like ``keops/clip.py``)

"""

import click

from .erase import erase
from .size import size
from .info import info
from .shrink import shrink
from .debug import debug


@click.group()
def main_group():
    """Keops command line interface"""
    pass


# In order to add commands to the group, just import
# and add them to the commands list
commands = [erase, size, info, shrink, debug]

for command in commands:
    main_group.add_command(command)



"""Drop a tile in a MBTiles file"""

import click

from .src.mvt_eraser import MVTEraser


@click.command(short_help='Erase a tile in a MBTiles file')
@click.argument('mbtiles', type=click.Path(exists=True), required=True)
@click.argument('zxy', type=str, required=True)
def erase(mbtiles: str, zxy: str):
    """Erase a tile in a MBTiles file.

    $ keops erase input.mbtiles 10/56/65

    """
    mvt_eraser = MVTEraser(mbtiles)
    mvt_eraser.erase_tile(zxy)



from pathlib import Path
from pytest import approx

from keops.src.mvt_reader import MVTReader

MBTILES = 'tests/fixtures/sample.mbtiles'
MBTILES_2 = 'tests/fixtures/trails.mbtiles'


def test_check_tile_exists():
    """

    :return:
    """
    reader = MVTReader(MBTILES)
    tile_exists = reader._check_tile_exists(8, 128, 160)

    assert tile_exists is True


def test_check_zoom_exists():
    """

    :return:
    """
    reader = MVTReader(MBTILES)
    zoom_exists = reader._check_zoom_exists(16)

    assert zoom_exists is False


def test_get_tile_size():
    """

    :return:
    """
    reader = MVTReader(MBTILES)
    tile_size = reader.get_tile_size('8/128/160')

    assert tile_size == approx(152.888)


def test_get_zoom_size():
    """

    :return:
    """
    reader = MVTReader(MBTILES)
    zoom_size = reader.get_zoom_size(8)

    assert zoom_size == approx(696.647)


def test_get_metadata():
    """

    :return:
    """
    reader = MVTReader(MBTILES_2)
    metadata = reader.get_metadata()
    expected = [('center', '-113.8348,48.6402,9'), ('format', 'pbf'), ('maxzoom', '14')]
    result = [i for i in metadata[0:3]]

    assert expected == result



from keops.src.utils import *


def test_decode_zxy_string():
    """
    Test the decode_zxy_string function from keops.src.utils
    """
    assert decode_zxy_string('6/10/56') == (6, 10, 56)


def test_zxy_string_is_valid():
    """
    Test the zxy_string_is_valid function from keops.src.utils
    """
    is_valid = zxy_string_is_valid('a/2/12')
    assert is_valid is False


def test_tile_zoom_are_valid():
    """
    Test the tile_zoom_are_valid function from keops.src.utils
    """
    is_valid = tile_zoom_are_valid(None, 'a/2/12')
    assert is_valid is False


def test_get_shrink_command_options():
    """
    Test the get_shrink_command_options function from keops.src.utils
    """
    expected = '--precision 2 --include 10'
    options = (None, 2, None, 10)
    result = get_shrink_command_options(options)
    assert expected == result


def test_mbtiles_is_valid():
    """
    Test the mbtiles_is_valid function from keops.src.utils
    """
    is_valid = mbtiles_is_valid('test/fixtures/empty.mbtiles')
    assert is_valid is False



"""Extract and print the metadata info from a MBTiles file"""

import click
import json

from .src.mvt_reader import MVTReader
from .src.utils import mbtiles_is_valid


@click.command(short_help='Extract and print the metadata info from a MBTiles file')
@click.argument('mbtiles', type=click.Path(exists=True), required=True)
@click.option('--show-json', is_flag=True, show_default=True, default=False, help='Show the JSON value [deault: False]')
def info(mbtiles: str, show_json: bool = False):
    """Extract and print the metadata info from a MBTiles file.

    $ keops info input.mbtiles

    $ keops info --show-json input.mbtiles

    """
    if not mbtiles_is_valid(mbtiles):
        return

    mvt_reader = MVTReader(mbtiles)

    metadata = mvt_reader.get_metadata()
    for name, value in metadata:
        if name != 'json':
            click.echo(f'{name}: {value}')
        else:
            if show_json:
                value = json.loads(value)
                click.echo(f'{name}: {json.dumps(value, sort_keys=True, indent=4, ensure_ascii=False)}')
            else:
                try:
                    json.loads(value)
                    click.echo(f'{name}: The value is a valid JSON, use --show-json for dump')
                except ValueError:
                    click.echo(f'{name}: The value is not a valid JSON')



import click
import re

from halo import Halo

GOOGLE_MERCATOR = 'EPSG:3857'


def decode_zxy_string(tile: str) -> tuple:
    """
    Decode a string with format z/x/y and return every parameter as an integer

    :param tile: tile position, as z/x/y
    :return: decoded tile position, with every parameter as an integer
    """
    zxy = tile.split('/')
    z, x, y = int(zxy[0]), int(zxy[1]), int(zxy[2])
    return z, x, y


def zxy_string_is_valid(tile: str) -> bool:
    """
    Check if a given tile position, with format z/x/y, is valid

    :param tile: tile position, as z/x/y
    :return: boolean that indicates if the given tile position is valid
    """
    zxy = tile.split('/')
    if len(zxy) != 3:
        click.echo(f'Tile format {tile} is not valid. It must follows the format z/x/y')
        return False
    for i in zxy:
        if re.search('[a-zA-Z]', i) is not None:
            click.echo('The given tile contains letters. It must follows the format z/x/y ONLY with integers')
            return False
    return True


def tile_zoom_are_valid(zoom: int or None, tile: str or None) -> bool:
    """
    Check if the user has given a tile position or a zoom, but not both at the same time
    or none of them

    :param tile: tile position, as z/x/y
    :param zoom: zoom, as integer
    :return: boolean that indicates if the user has given the inputs correctly
    """
    if zoom is None and tile is None:
        click.echo('You have to give at least a tile or a zoom level')
        return False
    elif zoom is not None and tile is not None:
        click.echo('You only have to give a tile or a zoom level, not both of them')
        return False
    elif zoom is None and tile is not None:
        return True if zxy_string_is_valid(tile) else False
    elif zoom is not None and tile is None:
        return True if type(zoom) == int else False


def mbtiles_is_valid(mbtiles: str) -> bool:
    """
    Check if a given file is a MBTiles, which is SQLite database

    :param mbtiles: file to check if is a valid MBTiles
    :return: boolean that indicates if the file is a valid MBTiles
    """
    from os.path import isfile, getsize

    error_message = '[>] The given MBTiles is not a valid SQLite database'
    if not isfile(mbtiles):
        click.echo(error_message)
        return False
    if getsize(mbtiles) < 100:  # SQLite database file header is 100 bytes
        click.echo(error_message)
        return False
    with open(mbtiles, 'rb') as fd:
        header = fd.read(100)

    if header[:16] != b'SQLite format 3\x00':
        click.echo(error_message)
        return False

    return True


def get_shrink_command_options(options: tuple) -> str:
    """
    Wrap the options given to the shrink command into a single string with them,
    in order to pass it to the tileshrink Docker
    :return: Return a string containing all the given commands
    """
    command_options = ''
    options_flags = {0: '--extent', 1: '--precision', 2: '--shrink', 3: '--include'}
    for option in options:
        if option is not None:
            i = options.index(option)
            flag = options_flags[i]
            option_command = f'{flag} {option}'
            if not command_options:
                command_options = option_command
            else:
                command_options = f'{command_options} {option_command}'

    return command_options


def get_spinner(text: str) -> Halo:
    """
    Get a Halo spinner in 'dots' format. For Windows users, it defaults to 'line'.

    :param text: text to display while spinning
    :return: Halo spinner object
    """
    return Halo(text=text, spinner='dots', color='yellow')



import click
from prettytable import PrettyTable


class MVTPrinter:
    """Print the given info as a beautiful table"""

    def print(self, fields: list or tuple, rows: list or tuple):
        """
        Print the given info as a beautiful table
        :param: fields: list with the name of the fields to be printed in the table
        :param: rows: list with the values of every row in the table
        """
        table = PrettyTable()
        table.field_names = fields
        for row in rows:
            # Check if the lenght of the given rows and the given columns are the same
            if len(row) != len(fields):
                continue
            table.add_row(row)

        click.echo(table)



import click

from .mvt_reader import MVTReader


class MVTDebugger(MVTReader):
    """
    Debug a MBTiles file. Get info related with layers and their
    features of a given tile or zoom level in a MBTiles file
    """

    def __init__(self, mbtiles: str):
        super().__init__(mbtiles)
        self.layers_dict = {}

    @staticmethod
    def digest_decoded_tile_data(decoded_tile_data: dict) -> dict:
        """
        Digest the decoded tile data and return a digerible
        dictionary with it
        :param: decoded_tile: decoded vector tile
        :return: digested_tile_data: dictionary with the digested tile data
        """
        digested_tile_data = {}
        for layer, layer_data in decoded_tile_data.items():
            features = layer_data['features']
            n_features = len(features)
            n_vertices = 0
            for feature in features:
                n_vertices += len(feature['geometry']['coordinates'])
            layer_dict = {
                "n_features": n_features,
                "n_vertices": n_vertices
            }
            digested_tile_data[layer] = layer_dict

        return digested_tile_data

    def get_digested_layers_dict(self, decoded_tiles: list) -> list or None:
        """
        Get a dict with the number of features and vertices of every layer
        in the MBTiles file.
        :param decoded_tiles: list of lists of tiles with the zoom level, tile column, tile row and
        decoded tile data
        :return: dictionary with the number of features and vertices of every layer
        """
        if decoded_tiles:
            for tile in decoded_tiles:
                self.add_tile_layers_to_dict(tile)
            return self.layers_dict
        else:
            click.echo('No data returned')   # TODO
            return

    def add_tile_layers_to_dict(self, tile: list):
        """
        Add the layers and them features and vertices from the given tile
        to the dict that is going to be returned
        :param tile: list with the zoom level, tile column, tile row and
        decoded tile data
        """
        data = tile[3]
        digested_data = self.digest_decoded_tile_data(data)
        for layer, info in digested_data.items():
            if layer in self.layers_dict:
                # Sum the number of features and vertices to the existing ones
                # Get the existing data to sum to the new one
                existing_layer_data = self.layers_dict.get(layer)
                existing_n_features = existing_layer_data['n_features']
                existing_n_vertices = existing_layer_data['n_vertices']
                # Get the new data
                new_n_features = existing_n_features + info['n_features']
                new_n_vertices = existing_n_vertices + info['n_vertices']
                # Update the key
                self.layers_dict[layer] = {'n_features': new_n_features, 'n_vertices': new_n_vertices}
            else:
                # Set the new layer as features and vertices
                self.layers_dict[layer] = {'n_features': info['n_features'], 'n_vertices': info['n_vertices']}



"""Reduce and simplify all features of all or any vector tiles in an MBTiles container"""

"""
Disclaimer

The merit of this module belongs entirely to @rastapasta, as the unique developer of tileshrink,
and @ooZberg, as the person who wrapped it in a Docker image in order to use it without worrying
about the Node.js version. What I have done is creating a backup of the Docker image and 
wrapping it again in this package, so it can be used in a focused environment.

Check them GitHubs here and give kudos to them!
@rastapasta: https://github.com/rastapasta
@ooZberg: https://github.com/ooZberg


Important

As this function uses a Docker image, you must have a running Docker version if you to use it. Also,
both the input and output MBTiles must be in the same directory, as it has to be mounted to
run the Docker image

"""


import os
import click
import docker

from .src.utils import get_shrink_command_options


@click.command(short_help='Reduce and simplify all features of all or any vector tiles in a MBTiles container. Docker required.')
@click.argument('input', type=click.Path(exists=True), required=True)
@click.argument('output', type=click.Path(), required=True)
@click.option('--extent', type=int, help='desired extent of the new layers [deault: 1024]')
@click.option('--precision', type=float, help='affects the level of simplification [deault: 1]')
@click.option('--shrink', type=int, help='maximal zoomlevel to apply the shrinking')
@click.option('--include', type=int, help='maximal zoomlevel to import untouched layers from [optional]')
def shrink(input: str, output: str, extent: int, precision: float, shrink: int, include: int):
    """Reduce and simplify all features of all or any vector tiles in a MBTiles container.

    Important:
        1. Docker required.
        2. Both input and output files must be in the same directory, as it has to be mounted
           to run Docker.

    $ keops shrink input.mbtiles output.mbtiles

    $ keops shrink --shrink 4 input.mbtiles output.mbtiles

    """
    # Input and output files must be different
    if input == output:
        click.echo('You must declare different input and output files')
        return

    # Get command options as string
    options = (extent, precision, shrink, include)
    command_options = get_shrink_command_options(options)

    # Get the input and output abspaths
    input_dir, output_dir = os.path.abspath(os.path.dirname(input)), os.path.abspath(os.path.dirname(output))
    if input_dir != output_dir:
        click.echo('Both input and output MBTiles must be in the same directory')
        return

    # Get the input and output file names
    input_file, output_file = os.path.basename(input), os.path.basename(output)
    # By default, the bind is /opt/data
    shrink_command = f'tileshrink {command_options} /opt/data/{input_file} /opt/data/{output_file}'
    shrink_mount = {input_dir: {'bind': '/opt/data', 'mode': 'rw'}}

    # Run the tileshrink
    client = docker.from_env()
    container_name = 'keops-tileshrink'
    image_name = "franmartin/tileshrink:latest"
    container = client.containers.run(image_name,
                                      name=container_name,
                                      volumes=shrink_mount,
                                      command=shrink_command,
                                      detach=True,
                                      remove=True)

    logs = container.logs(stream=True)
    for line in logs:
        click.echo(line.strip())



import sqlite3
import gzip
import mapbox_vector_tile
import click

from .utils import decode_zxy_string


TILE_SIZE_LIMIT = 500


class MVTReader:
    """Read a MBTiles file and return its data, decoded or not"""

    def __init__(self, mbtiles: str):
        self.mbtiles = mbtiles
        self.conn = self._get_conn()
        self.cur = self._get_cur()

    def _get_conn(self) -> sqlite3.connect:
        """
        Create the SQLite connection

        :return: SQlite connection
        """
        conn = self._create_connection(self.mbtiles)
        return conn

    def _get_cur(self) -> sqlite3.Cursor:
        """
        Get a SQLite connection cursor

        :return: SQLite cursor
        """
        return self.conn.cursor()

    def _query(self, sql_query: str, rows: bool = False, index: int = -1) -> list or None:
        """
        Run a given SQL query in the MBTiles file

        :param sql_query: SQL query to run in the MBTiles file
        :param rows: boolean that indicates if the response must be a list with all the returned rows
        :param index: integer that indicates which of the columns from the response have to be returned
        :return: query response from the MBTiles file
        """
        try:
            if rows:
                if index >= 0:
                    self.cur.row_factory = lambda cursor, row: row[index]
                elif index < 0:
                    self.cur.row_factory = lambda cursor, row: row
            self.cur.execute(sql_query)
            response = self.cur.fetchall()
            self.cur.row_factory = None
            self.conn.close()
        except Exception as e:
            click.echo(e)
            return

        return response

    def get_tile_size(self, zxy: str) -> float:
        """
        Get the size in KB of a given tile in the MBTiles

        :param zxy: tile position, as z/x/y
        :return: size of the given tile, as KB
        """
        z, x, y = decode_zxy_string(zxy)
        tile_exists = self._check_tile_exists(z, x, y)

        if not tile_exists:
            self.conn.close()
            click.echo(f'Error: The given tile at {zxy} does not exists in the MBTiles file')
            return 0

        # Get the size in KB
        query = f'SELECT length(tile_data) as size FROM tiles WHERE zoom_level={z} AND tile_column={x} AND tile_row={y}'
        tile_size = self._query(query, True, 0)[0] * 0.001
        tile_size = round(tile_size, 3)

        return tile_size

    def get_zoom_size(self, z: int) -> float:
        """
        Get the size in KB of a given zoom in the MBTiles

        :param z: zoom to get the zoom of
        :return: size of the given zoom, as KB
        """
        zoom_exists = self._check_zoom_exists(z)

        if not zoom_exists:
            self.conn.close()
            click.echo(f'Error: The given zoom {z} does not exists in the MBTiles file')
            return 0

        # Sum the response and get the size in KB
        query = f'SELECT length(tile_data) as size FROM tiles WHERE zoom_level={z}'
        zoom_size = sum(self._query(query, True, 0)) * 0.001
        zoom_size = round(zoom_size, 3)

        return zoom_size

    def get_tile(self, zxy: str) -> list or None:
        """
        Get the data associated to a given tile, such as its zoom level, tile column, tile row
        and the encoded tile data as a Google proto buff

        :param zxy: tile position, as z/x/y
        :return: data associated to a given tile, with the encoded tile data
        """
        z, x, y = decode_zxy_string(zxy)
        tile_exists = self._check_tile_exists(z, x, y)

        if not tile_exists:
            self.conn.close()
            click.echo(f'Error: The given tile at {zxy} does not exists in the MBTiles file')
            return

        query = f'SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles WHERE zoom_level={z} AND tile_column={x} AND tile_row={y}'
        tile = self._query(query)[0]

        return tile

    def get_decoded_tile(self, zxy: str) -> list or None:
        """
        Get the data associated to a given tile, such as its zoom level, tile column, tile row
        and the decoded tile data as geoJSON

        :param zxy: tile position, as z/x/y
        :return: data associated to a given tile, with the decoded tile data
        """
        z, x, y = decode_zxy_string(zxy)
        tile_exists = self._check_tile_exists(z, x, y)

        if not tile_exists:
            self.conn.close()
            click.echo(f'Error: The given tile at {zxy} does not exists in the MBTiles file')
            return

        query = f'SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles WHERE zoom_level={z} AND tile_column={x} AND tile_row={y}'
        tile = self._query(query)[0]
        decoded_tile_data = self._decode_tile_data(tile)
        decoded_tile = [tile[0], tile[1], tile[2], decoded_tile_data]

        return decoded_tile

    def get_decoded_zoom(self, z: int) -> list or None:
        """
        Get the data associated to all the tiles in a given tile, such as them zoom level, tile column, tile row,
        and the decoded tile data as geoJSON

        :param z: zoom to get the data of
        :return: data associated to all the tiles in a given zoom, with the decoded tile data
        """
        zoom_exists = self._check_zoom_exists(z)

        if not zoom_exists:
            self.conn.close()
            click.echo(f'Error: The given zoom {z} does not exists in the MBTiles file')
            return

        # Get the decoded tiles from a given zoom
        query = f'SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles WHERE zoom_level={z}'
        zoom_tiles = self._query(query, True)
        decoded_zoom_data = []
        for tile in zoom_tiles:
            decoded_tile_data = self._decode_tile_data(tile)
            decoded_tile = [tile[0], tile[1], tile[2], decoded_tile_data]
            decoded_zoom_data.append(decoded_tile)

        return decoded_zoom_data

    def get_tiles(self) -> list or None:
        """
        Query and return the features of all the tiles in the MBTiles

        Estructure of the tiles
        for tile in tiles:
            tile[0] = zoom
            tile[1] = x column
            tile[2] = y column
            tile[3] = encoded data

        :return: tiles: tuple containing data in the MBTiles
        """
        query = 'SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles'

        tiles = self._query(query)

        if tiles:
            return tiles
        else:
            click.echo('There is no data in the MBTiles')
            return

    def get_big_tiles(self) -> list or None:
        """
        Query and return the features of the big tiles in the MBTiles

        Estructure of the tiles
        for tile in tiles:
            tile[0] = zoom
            tile[1] = x column
            tile[2] = y column
            tile[3] = encoded data

        :return: tiles: tuple containing data in the MBTiles
        """
        query = f'SELECT zoom_level, tile_column, tile_row, tile_data, length(tile_data) as size FROM tiles WHERE length(tile_data) > {TILE_SIZE_LIMIT * 1024} ORDER BY zoom_level, tile_column, tile_row ASC'

        tiles = self._query(query, True)

        if tiles:
            return tiles
        else:
            click.echo('There is no data or at least not too big tiles in the MBTiles')
            return

    def get_decoded_tiles(self, size_limit=False) -> list or None:
        """
        Query, decode and return the features of the MBTiles

        Estructure of the decoded tiles
        for tile in tiles:
            tile[0] = zoom
            tile[1] = x column
            tile[2] = y column
            tile[3] = decoded data
            for layer, layer_data in tile[3].items():
                layer = layer name
                layer_data = geojson

        :param: size_limit: indicates if the there is a size constraint
        in the query
        :return: tiles: tuple containing decoded data in the MBTiles
        """
        decoded_tiles = []
        tiles = self.get_tiles() if not size_limit else self.get_big_tiles()

        if tiles:
            for tile in tiles:
                decoded_tile_data = self._decode_tile_data(tile)
                decoded_tile = [tile[0], tile[1], tile[2], decoded_tile_data]
                decoded_tiles.append(decoded_tile)
        else:
            click.echo('There is no data or at least not too big tiles in the MBTiles')
            return

        return decoded_tiles

    def get_metadata(self) -> list or None:
        """
        Get the info from the metadata table in a MBTiles file

        :return: MBTiles metadata
        """
        query = 'SELECT name, value FROM metadata'

        metadata = self._query(query)

        if metadata:
            return metadata
        else:
            click.echo('There is no metadata in the MBTiles')
            return

    @staticmethod
    def _create_connection(db_file: str) -> sqlite3.connect:
        """
        Create a database connection to the SQLite database
        specified by the db_file

        :param db_file: database file
        :return: connection object or None
        """
        conn = None
        try:
            conn = sqlite3.connect(db_file)
        except Exception as e:
            click.echo(e)

        return conn

    @staticmethod
    def _get_tile_zxy(tile: tuple) -> tuple:
        """
        Get the zoom level, tile column and 
        tile row

        :param: tile: vector tile object
        :return: zoom level, tile column and tile row
        """
        zxy = (tile[0], tile[1], tile[2])
        return zxy

    @staticmethod
    def _get_tile_zxy_string(tile: tuple) -> str:
        """
        Get the zoom level, tile column and tile row
        in string format

        :param: tile: vector tile object
        :return: string with the zoom level, tile column
        and tile row
        """
        zxy = f'{tile[0]}/{tile[1]}/{tile[2]}'
        return zxy

    @staticmethod
    def _get_tile_data(tile: tuple) -> str:
        """
        Get decompressed tile data

        :param: tile: vector tile object
        :return: decompressed tile data
        """
        return gzip.decompress(tile[3])

    def _decode_tile_data(self, tile: tuple) -> dict:
        """
        Get the decoded tile data

        :param: tile_data: data in the vector tile
        :return: decoded tile data
        """
        encoded_tile_data = self._get_tile_data(tile)
        try:
            return mapbox_vector_tile.decode(encoded_tile_data)
        except Exception as e:
            click.echo(e)

    def _check_tile_exists(self, z: int, x: int, y: int) -> bool:
        """
        Check if a given tile exists in the MBTiles file

        :param: z: zoom level of the given tile
        :param: x: tile column of the given tile
        :param: y: tile row of the given tile
        :return: boolean that indicates if the given tile exists or not
        """
        query = f'SELECT * from tiles WHERE zoom_level={z} AND tile_column={x} AND tile_row={y};'

        try:
            self.cur.execute(query)
            result = self.cur.fetchone()
            if result:
                return True
            else:
                return False
        except Exception as e:
            click.echo(e)
            return False

    def _check_zoom_exists(self, z: int) -> bool:
        """
        Check if a given zoom exists in a MBTiles file

        :param z: zoom level to get if exists or not
        :return: boolean that indicates if the given zoom exists or not
        """
        query = f'SELECT * from tiles WHERE zoom_level={z} LIMIT 1'

        try:
            self.cur.execute(query)
            result = self.cur.fetchone()
            if result:
                return True
            else:
                return False
        except Exception as e:
            click.echo(e)
            return False



"""Debug a MBTiles file: get info related with layers and their features in a given MBTiles"""

import click

from .src.mvt_printer import MVTPrinter
from .src.mvt_debugger import MVTDebugger
from .src.utils import mbtiles_is_valid, get_spinner


@click.command(short_help='Debug a MBTiles file: get info related with layers and their features in a given MBTiles')
@click.argument('mbtiles', type=click.Path(exists=True), required=True)
@click.option('-z', '--zoom', type=int, help='Zoom of the MBTiles to get the size of')
@click.option('-t', '--tile', type=str, help='Tile of the MBTiles to get the size of')
def debug(mbtiles: str, zoom: int, tile: str):
    """Debug a MBTiles file: get info related with layers and their features in a given MBTiles.

    $ keops debug --zoom 10 input.mbtiles

    $ keops debug --tile 10/56/65 input.mbtiles

    """
    if not mbtiles_is_valid(mbtiles):
        return

    mvt_printer = MVTPrinter()
    mvt_debugger = MVTDebugger(mbtiles)
    decoded, layers_dict = None, None

    spinner = get_spinner('Debugging the MBTiles...')
    spinner.start()
    if tile is None and zoom is None:
        # Debug the entire MBTiles
        decoded = mvt_debugger.get_decoded_tiles()
        layers_dict = mvt_debugger.get_digested_layers_dict(decoded)
    elif tile is not None and zoom is None:
        # Debug a tile
        decoded = mvt_debugger.get_decoded_tile(tile)
        mvt_debugger.add_tile_layers_to_dict(decoded)
        layers_dict = mvt_debugger.layers_dict
    elif tile is None and zoom is not None:
        # Debug a zoom
        decoded = mvt_debugger.get_decoded_zoom(zoom)
        layers_dict = mvt_debugger.get_digested_layers_dict(decoded)

    spinner.stop()
    if layers_dict:
        rows = []
        for layer, info in layers_dict.items():
            f, v = info['n_features'], info['n_vertices']
            row = (layer, f, v)
            rows.append(row)

        fields = ('layer', 'n_features', 'n_vertices')
        mvt_printer.print(fields, rows)



import shutil
import os

from keops.src.mvt_eraser import MVTEraser


def test_mvt_eraser_tiles():
    """

    :return:
    """
    src_mbtiles, tmp_mbtiles = 'tests/fixtures/sample.mbtiles', 'tests/fixtures/sample-temp.mbtiles'
    # Create a temporal copied version of the sample MBTiles files, to avoid deleting
    # real data
    shutil.copyfile(src_mbtiles, tmp_mbtiles) if not os.path.isfile(tmp_mbtiles) else None

    mvt_eraser = MVTEraser(tmp_mbtiles)
    mvt_eraser.erase_tile('8/128/160')
    tile_exists = mvt_eraser._check_tile_exists('8', '128', '160')

    # Remove the temporal MBTiles file
    os.remove(tmp_mbtiles)

    assert tile_exists is False


def test_mvt_eraser_map_images():
    """

    :return:
    """
    src_mbtiles, tmp_mbtiles = 'tests/fixtures/trails.mbtiles', 'tests/fixtures/sample-temp.mbtiles'
    # Create a temporal copied version of the sample MBTiles files, to avoid deleting
    # real data
    shutil.copyfile(src_mbtiles, tmp_mbtiles) if not os.path.isfile(tmp_mbtiles) else None

    mvt_eraser = MVTEraser(tmp_mbtiles)
    mvt_eraser.erase_tile('8/128/160')
    tile_exists = mvt_eraser._check_tile_exists('8', '46', '168')

    # Remove the temporal MBTiles file
    os.remove(tmp_mbtiles)

    assert tile_exists is False


def test_get_tiles_table():
    """

    :return:
    """
    mbtiles = 'tests/fixtures/sample.mbtiles'

    mvt_eraser = MVTEraser(mbtiles)
    result = mvt_eraser._get_tiles_table()
    expected = 'tiles'

    assert expected == result



import click

from .mvt_reader import MVTReader
from .utils import decode_zxy_string


class MVTEraser(MVTReader):
    """Erase a tile in a MBTiles file"""

    def __init__(self, mbtiles: str):
        super().__init__(mbtiles)
        self.tile_table = self._get_tiles_table()

    def _get_tiles_table(self) -> str or None:
        """
        Check which is the table with the tile data. Depending on the MBTiles provider, it can be
        'tiles' or 'map'
        :return: name of the table that contains the tile data
        """
        query = "SELECT name FROM sqlite_master WHERE type='table';"
        self.cur.row_factory = lambda cursor, row: row[0]

        try:
            self.cur.execute(query)
            result = self.cur.fetchall()
            self.cur.row_factory = None

            if 'tiles' in result:
                return 'tiles'
            elif 'map' in result and 'images' in result:
                return 'map'
            else:
                click.echo('The are no tiles table in the MBTiles')
                return False
        except Exception as e:
            click.echo(e)
            self.cur = None
            return False

    def _erase_tile_from_tiles(self, z: int, x: int, y: int) -> None:
        """
        Erase a given tile from the tiles table
        """
        query = f'DELETE FROM tiles WHERE zoom_level={z} AND tile_column={x} AND tile_row={y}'

        try:
            self.cur.execute(query)
            click.echo(f'[>] Tile in {z}/{x}/{y} erased successfully!')
            self.conn.close()
        except Exception as e:
            click.echo(e)
            return

    def _erase_tile_from_map_images(self, z: int, x: int, y: int) -> None:
        """
        Erase given tile from the map table
        :param: z: Zoom level of the given tile
        :param: x: Tile column of the given tile
        :param: y: Tile row of the given tile
        """
        try:
            query = f'SELECT tile_id FROM map WHERE zoom_level={z} AND tile_column={x} AND tile_row={y};'
            self.cur.execute(query)
            tile_id_ = self.cur.fetchone()
            tile_id = tile_id_[0]
        except Exception as e:
            click.echo(e)
            return

        script = f"""
            DELETE FROM map WHERE tile_id='{tile_id}';
            DELETE FROM images WHERE tile_id='{tile_id}';
        """
        self.cur.executescript(script)
        click.echo(f'[>] Tile in {z}/{x}/{y} erased successfully!')

    def erase_tile(self, zxy: str):
        """
        Erase a given tile in a MBTiles file
        :param zxy: Tile position, as z/x/y
        """
        if not self.tile_table:
            return False

        z, x, y = decode_zxy_string(zxy)
        tile_exists = self._check_tile_exists(z, x, y)
        if not tile_exists:
            self.conn.close()
            click.echo(f'Error: The given tile at {zxy} does not exists in the MBTiles file')
            return False

        if self.tile_table == 'tiles':
            self._erase_tile_from_tiles(z, x, y)
        elif self.tile_table == 'map':
            self._erase_tile_from_map_images(z, x, y)



"""Get the size in KB of a given tile or given zoom level"""

import click

from .src.mvt_reader import MVTReader
from .src.utils import tile_zoom_are_valid, mbtiles_is_valid


@click.command(short_help='Get the size of a given tile or zoom level in a MBTiles file')
@click.argument('mbtiles', type=click.Path(exists=True), required=True)
@click.option('-z', '--zoom', type=int, help='Zoom of the MBTiles to get the size of')
@click.option('-t', '--tile', type=str, help='Tile of the MBTiles to get the size of')
def size(mbtiles: str, zoom: int, tile: str):
    """Get the size in KB of a given tile or given zoom level.

    $ keops size --zoom 10 input.mbtiles

    $ keops size --tile 10/56/65 input.mbtiles

    """
    if not tile_zoom_are_valid(zoom, tile) and not mbtiles_is_valid(mbtiles):
        return

    mvt_reader = MVTReader(mbtiles)
    if tile is not None and zoom is None:
        tile_size = mvt_reader.get_tile_size(tile)
        if tile_size:
            click.echo(f'[>] The size of the tile {tile} is {tile_size} KB')
            if tile_size > 500:
                click.echo(f'Remember that the maximum recommended size of a tile is 500 KB!')
    elif tile is None and zoom is not None:
        zoom_size = mvt_reader.get_zoom_size(zoom)
        if zoom_size:
            click.echo(f'[>] The size of the zoom {zoom} is {zoom_size} KB')
